/**
 * Defines a doubly-linked list class
 * @author
 * @author
 */
import java.util.NoSuchElementException;
import java.lang.SuppressWarnings;

/**
 * LinkedList.java
 * @author jenniferparrish
 * @author Osbert Sudjana
 * @author Sharon Utomo
 * CIS 22C, Lab 5
 */

public class LinkedList<T> {
	private class Node {
		private T data;
	    private Node next;
	    private Node prev;

	    public Node(T data) {
	    	this.data = data;
	        this.next = null;
	        this.prev = null;
	    }
	}
	
	private int length;
	private Node first;
	private Node last;
	private Node iterator;

		/**** CONSTRUCTORS ****/

		/**
	 	* Instantiates a new LinkedList with default values
     	* @postcondition Instantiates a new linked list object
     	*/
        public LinkedList() {
        	first = null;
        	last = null;
        	iterator = null;
        	length = 0;
        }

	    /**
	     * Converts the given array into a LinkedList
	     * @param array the array of values to insert into this LinkedList
	     * @postcondition Instantiates a new linked list using data from array
	     */
	    public LinkedList(T[] array) {
	    	if(array == null) {
	    		first = null;
	        	last = null;
	        	iterator = null;
	        	length = 0;
	    	}
	    	else {
	    		for (int i = 0; i < array.length; i++) {
	    			addLast(array[i]);
	    		}
	    	}
	    }
	    
	    /**
	     * Instantiates a new LinkedList by copying another List
	     * @param original the LinkedList to copy
	     * @postcondition a new List object, which is an identical,
	     * but separate, copy of the LinkedList original
	     */
	    public LinkedList(LinkedList<T> original) {
	        if (original == null) {
	            return;
	        }
	        if (original.length == 0) {
	            length = 0;
	            first = null;
	            last = null;
	            iterator = null;
	        } else {
	            Node temp = original.first;
	            while (temp != null) {
	                this.addLast(temp.data);
	                temp = temp.next;
	            }
	            iterator = null;
	        }
	    }
	    

	    /**** ACCESSORS ****/

	    /**
	     * Returns the value stored in the first node
	     * @precondition the list is not empty
	     * @return the value stored at node first
	     * @throws NoSuchElementException <when the precondition is violated>
	     */
	    public T getFirst() throws NoSuchElementException {
	        if (length == 0) {
	        	throw new NoSuchElementException("getFirst: "
	        			+ "List is empty. No data to access!");
	        }
	        return first.data;
	    }

	    /**
	     * Returns the value stored in the last node
	     * @precondition the list is not empty
	     * @return the value stored in the node last
	     * @throws NoSuchElementException when the precondition is violated
	     */
	    public T getLast() throws NoSuchElementException {
	    	if (length == 0) {
	    		throw new NoSuchElementException("getLast: "
		        		+ "List is empty. No data to access!");
		    }
		    return last.data;
	    }

	    /**
	     * Returns the data stored in the iterator node
	     * @precondition iterator is not offEnd
	     * @throw NullPointerException when iterator is null
	     */
	    public T getIterator() throws NullPointerException {
	    	if (offEnd()) {
	    		throw new NullPointerException("getIterator: iterator is null!");
	    	}
	        return iterator.data;
	    }

	    /**
	     * Returns the current length of the LinkedList
	     * @return the length of the LinkedList from 0 to n
	     */
	    public int getLength() {
	        return length;
	    }

		/**
	     * Returns whether the LinkedList is currently empty
	     * @return whether the LinkedList is empty
	     */
	    public boolean isEmpty() {
	    	return length == 0;
	    }

	 

	    /**
	     * Returns whether the iterator is offEnd, i.e. null
	     * @return whether the iterator is null
	     */
	     public boolean offEnd() {
	         return iterator == null;
	     }

	 

	    /**** MUTATORS ****/

	    /**
	     * Creates a new first element
	     * @param data the data to insert at the front of the LinkedList
	     * @postcondition creates a new first node
	     */
	    public void addFirst(T data) {
	    	Node n = new Node (data);
	        if (length == 0) { //edge case
	        	first = last = n;
	        }
	        else {
	        	first.prev = n;
	        	n.next = first;
	        	first = n;
	        }
	        length++;
	    }

	    /**
	     * Creates a new last element
	     * @param data the data to insert at the end of the LinkedList
	     * @postcondition creates a new last node
	     */
	    public void addLast(T data) {
	        Node n = new Node (data);
	        if (length == 0) { //edge case
	        	first = last = n;
	        }
	        else {
	        	last.next = n;
	        	n.prev = last;
	        	last = n;
	        }
	        length++;
	    }
	    
	    /**
	     * Inserts a new element after the iterator
	     * @param data the data to insert
	     * @precondition iterator is not offEnd
	     * @throws NullPointerException if the iterator is offEnd
	     */
	    public void addIterator(T data) throws NullPointerException{
	    	if(offEnd()) { //precondition
	    		throw new NullPointerException ("addIterator: Iterator is offend. Cannot add");
	    	}
	    	else if(iterator == last) { //edge case
	    		addLast(data);
	    	}
	    	else { //general case
	        Node n = new Node (data);
	        n.prev = iterator;
	        n.next = iterator.next;
	        iterator.next.prev = n;
	        iterator.next = n;  
	        length++;
	    	}
	    }

	    /**
	     * removes the element at the front of the LinkedList
	     * @precondition list is not empty
	     * @postcondition node at the front of the linked list is removed. 
	               If linked list only has 1 node, linked list is emptied
	     * @throws NoSuchElementException if linked list is empty
	    */
	    public void removeFirst() throws NoSuchElementException{
	        if (length == 0) { //precondition
	            throw new NoSuchElementException("removeFirst(): Cannot remove from an empty List!");
	        } else if (length == 1) { //edge case #1
	            first = last = iterator = null;
	        } else {
	            if(iterator == first) { //edge case #2
	                iterator = null;
	            }
	            //general case:
	            first = first.next;
	            first.prev = null;
	        }
	        length--;
	    }


	    /**
	     * removes the element at the front of the LinkedList
	     * @precondition list is not empty
	     * @postcondition node at the end of the linked list is removed. 
	              If linked list only has 1 node, linked list is emptied
	     * @throws NoSuchElementException if linked list is empty
	    */
	    public void removeLast() throws NoSuchElementException {
	    	if (length == 0) { //precondition
	            throw new NoSuchElementException("removeFirst(): Cannot remove from an empty List!");
	        } else if (length == 1) { //edge case #1
	            first = last = iterator = null;
	        } else {
	            if(iterator == last) { //edge case #2
	                iterator = null;
	            }
	            //general case:
	            last = last.prev;
	            last.next = null;
	        }
	        length--;
	    }
	    
	    /**
	     * removes the element referenced by the iterator
	     * @precondition iterator != null
	     * @postcondition element referenced by iterator is removed
	     * @throws NullPointerException if iterator is offEnd
	     */
	    public void removeIterator() throws NullPointerException {
	        if(offEnd()) { //precondition
	        	throw new NullPointerException("removeIterator: Iterator is offEnd. Cannot remove");
	        }
	        else if(iterator == first) { //edge case 1
	        	removeFirst();
	        }
	        else if(iterator == last) { //edge case 2
	        	removeLast();
	        }
	        else { //general case
	            Node temp = iterator;
	            Node temp2 = iterator.next;
	            length--;
	            iterator = null;
	            temp.prev.next = temp2;
	            temp2.prev = temp.prev;
	        }
	    }
	    
	    /**
	     * places the iterator at the first node
	     * @postcondition moves iterator to the beginning of the list
	     */
	    public void positionIterator(){
	        iterator = first;
	    }
	    
	    /**
	     * Moves the iterator one node towards the last
	     * @precondition iterator is not offEnd
	     * @postcondition iterator is advanced one node forward
	     * @throws NullPointerException when iterator is offEnd
	     */
	    public void advanceIterator() throws NullPointerException {
	    	if (offEnd()) {
	    		throw new NullPointerException ("advanceIterator: "
	    				    + "iterator is null. Cannot advance!");
	    	}
	        iterator = iterator.next;
	    }
	    
	    /**
	     * Moves the iterator one node towards the first
	     * @precondition iterator is not offEnd
	     * @postcondition iterator is moved one node backward
	     * @throws NullPointerException if iterator is offEnd
	     */
	    public void reverseIterator() throws NullPointerException {
	    	if (offEnd()) {
	    		throw new NullPointerException ("advanceIterator: "
	    				    + "iterator is null. Cannot advance!");
	    	}
	        iterator = iterator.prev;
	    }

	    /**** ADDITIONAL OPERATIONS ****/

	    /**
	     * Converts the LinkedList to a String, 
	     * with each value separated by a blank
	     * line At the end of the String, place a 
	     * new line character
	     * @return the LinkedList as a String
	     */
	    @Override public String toString() {
	        StringBuilder result = new StringBuilder();
	        Node temp = first;
	        while(temp != null) {
	            result.append(temp.data + " ");
	            temp = temp.next;
	        }
	        return result.toString() + "\n";
	    }
	    
	    /**
	     * Determines whether the given Object is
	     * another LinkedList, containing
	     * the same data in the same order
	     * @param o another Object
	     * @return whether there is equality
	     */
	    @SuppressWarnings("unchecked") //good practice to remove warning here
	    @Override 
	    public boolean equals(Object o) {
	        if(o == this) {
	            return true;
	        } else if (!(o instanceof LinkedList)) {
	            return false;
	        } else {
	            LinkedList<T> L = (LinkedList<T>) o; 
	            if (this.length != L.length) {
	                return false;
	            } else {
	                Node temp1 = this.first;
	                Node temp2 = L.first;
	                while (temp1 != null) {
	                	if (!(temp1.data.equals(temp2.data))) {
	                		return false;
	                	}
	                	temp1 = temp1.next;
	                	temp2 = temp2.next;
	                }
	            }
	                return true;
	        }
	     }
	    
	   
	    /**CHALLENGE METHODS*/
	    
	    
	    /**
	     * Zippers two LinkedLists to create a third List
	     * which contains alternating values from this list
	     * and the given parameter
	     * For example: [1,2,3] and [4,5,6] -> [1,4,2,5,3,6]
	     * For example: [1, 2, 3, 4] and [5, 6] -> [1, 5, 2, 6, 3, 4]
	     * For example: [1, 2] and [3, 4, 5, 6] -> [1, 3, 2, 4, 5, 6]
	     * @param list the second LinkedList in the zipper
	     * @return a new LinkedList, which is the result of
	     * zippering this and list
	     * @postcondition this and list are unchanged
	     */
	    public LinkedList<T> zipperLists(LinkedList<T> list) {
	    	
	    	LinkedList<T> newList = new LinkedList<T>();
	    	
	        Node temp1 = this.first;
	        Node temp2 = list.first;
	        
	        if (this.length == list.length) {
	        	while(temp1 != null) {
	        		newList.addLast(temp1.data);
	        		newList.addLast(temp2.data);
	        		temp1 = temp1.next;
	        		temp2 = temp2.next;
	        	}
	        }
	        
	        if (this.length > list.length) {
	        	while(temp2 != null) {
	        		newList.addLast(temp1.data);
	        		newList.addLast(temp2.data);
	        		temp1 = temp1.next;
	        		temp2 = temp2.next;
	        	}
	        	while(temp1 != null) {
	        		newList.addLast(temp1.data);
	        		temp1 = temp1.next;
	        	}
	        }
	        
	        if (this.length < list.length) {
	        	while(temp1 != null) {
	        		newList.addLast(temp1.data);
	        		newList.addLast(temp2.data);
	        		temp1 = temp1.next;
	        		temp2 = temp2.next;
	        	}
	        	while(temp2 != null) {
	        		newList.addLast(temp2.data);
	        		temp2 = temp2.next;
	        	}
	        }
	        
	        return newList;
	        
	    }
	    
	    /**
	     * Determines whether a LinkedList is
	     * reversible, i.e. the data is the same
	     * written both forward and backward
	     * e.g. isReversible(1 2 3 2 1) -> true
	     * e.g. isReversible(A B B A) -> true
	     * @return whether the list is reversible
	     */
	    public boolean isReversible() {
	    	Node n1 = this.first;
	    	Node n2 = this.last;
	    	if(length % 2 == 1) {
	    		while (n1 != n2) {
	    			if (!(n1.data.equals(n2.data))) {
	    				return false;
	    			}
	    			n1 = n1.next;
	    			n2 = n2.prev;
	    		}
	    	}
	    	else {
	    		for (int i = 0; i < length/2; i++) {
	    			if (!(n1.data.equals(n2.data))) {
	    				return false;
	    			}
	    			n1 = n1.next;
	    			n2 = n2.prev;
	    		}
	    	}
	        return true;
	    }
	    
	    /**
	     * Determines at which index the iterator is located
	     * Indexed from 0 to length - 1
	     * @return the index number of the iterator
	     * @precondition !offEnd()
	     * @throws NullPointerException when precondition is violated
	     */
	    public int getIteratorIndex() throws NullPointerException {
	        if(offEnd()) {
	        	throw new NullPointerException ("getIteratorindex(): iterator is offEnd. cannot get Index!");
	        }
	        Node temp = iterator;
	        int index = 0;
	        positionIterator();
	        while (iterator != temp) {
	        	advanceIterator();
	        	index++;
	        }
	        return index;
	    }
	    
	    /**
	     * Searches the LinkedList for a given
	     * element's index
	     * @param data the data whose index to locate
	     * @return the index of the data or -1 if
	     * the data is not contained in the LinkedList
	     */
	    public int findIndex(T data) {
	        positionIterator();
	        for (int i = 0; i < this.length; i++) {
	        	if (iterator.data.equals(data)) {
	        		iterator = null;
	        		return i;
	        	}
	        	advanceIterator();
	        }
	        iterator = null;
	        return -1;
	    }
	    
	    /**
	     * Advances the iterator to location within
	     * the LinkedList specified by the given index
	     * @param index the index at which to place the
	     * iterator.
	     * @precondition !offEnd()
	     * @throws NullPointerException when the precondition is
	     * violated
	     */
	    public void advanceIteratorToIndex(int index) throws NullPointerException {
	       if (offEnd()) {
	    	   throw new NullPointerException ("advanceIteratorToIndex(): Iterator is offEnd. Cannot advance!");
	       }
	       for (int i = 0; i < index; i++) {
	    	   advanceIterator();
	       }
	    }
}
	
